// Trinket control program for Pinscape fan topper lighting
//
// Buy a trinket: https://www.adafruit.com/product/1501
//
// You can compile this program through an Arduino IDE.
// See https://learn.adafruit.com/introducing-trinket/setting-up-with-arduino-ide
//
//
// My virtual pinball machine, Pinscape, has a fan on top of the backbox,
// with a custom enclosure that I designed and had 3D-printed.  The
// enclosure features two LED lighting features.  The first is a 5050
// RGB LED strip (the type you can find on eBay, commonly used on pin
// cabs for under-cab ambient lighting), wrapped around the inside of
// the front fan opening, facing inwards towards the fan rotation axis.
// The 5050 strip is wired in parallel with my under-cab light strips,
// so it simply replicates the under-cab light program, which is a
// standard DOF effect.  (I use the DOF "undercab complex" program,
// which flashes different colors in response to game events.)  This
// Trinket project doesn't have anything to do with the light strip,
// since those are simply wired into the DOF controller directly via
// the undercab-RGB ports.
//
// This project controls the second lighting feature on the fan, which
// is a ring of nine individual 3W RGB LEDs poking out through holes
// in the face of the enclosure, facing forwards towards the player.
// These LEDs are wired in series so that they all show the same color.
// These are wired to the Trinket via 3 GPIO ports (for the red, green,
// and blue channels).
//
// Why do we need the Trinket at all?  Because I wanted the perimeter
// LEDs to show a separate color pattern that's not in the standard
// DOF effects.  To be specific, I wanted the perimeter lights to
// flash white whenever the strobes fire, and to flash a complex
// red/blue/white pattern when the beacons fire.  I wanted to
// reproduce the kind of light show you see in the big crime scenes
// on the 11 o'clock news, with all of the police cruisers and
// ambulances flashing their colored strobe lights.  So that's what
// this program is all about: it monitors the strobe and beacon ports
// in the DOF controller, and when it sees them fire, it shows the
// effects I wanted on the fan LEDs.
//
// It's probably possible to accomplish substantially the same effect
// directly with DOF, without extra hardware, but the DOF Config tool
// just makes it too hard.  What you'd have to do is assign the fan
// LEDs as a "Custom RGB" device, then you'd have to create a program
// in the Config Tool for this custom device that uses the same
// triggers as the Strobe and Beacon devices, and shows the desired
// flash pattern when triggered.  The problem is that DOF doesn't have
// way to say "trigger Custom RGB when Strobe is triggerd"; it only has
// a way to say things like "trigger when Switch 25 in the game is
// triggered".  The triggers for the individual devices in the DOF
// programs tend to be quite complex, so you'd have to reiterate a big
// list of triggers for the Custom RGB device, and you'd have to repeat
// this process separately for EVERY SINGLE GAME, since every game has
// its own unique set of triggers for every device. The only way to
// make this remotely workable would be to do your own scripted post-
// processing of the Config Tool output to mechanically generate the
// derived effects.  The bottom line is that it seemed simpler to me to
// do it with a separate piece of hardware (and that's coming from a
// software person!).
//
// Due to the high current draw on the LEDs, we need some kind of
// amplifier/booster circuit between the Trinket and the LEDs.  For
// my setup, I use a MOSFET for each output, but any sort of transistor
// circuit that can switch about 500mA will work.  (The series wiring
// of the 3W LEDs means that each color channel draws about 350mA at
// full brightness.)
//
// On the input side, the Trinket is wired into the DOF controller
// *outputs*, for the strobes and beacons.  One Trinket GPIO port is
// connected to the DOF strobe output port, and one Trinket GPIO is
// wired to the DOF beacon output port.  You can think of the DOF
// output controller as a low-side transistor switch (which is
// literally how it's implemented for Pinscape, LedWiz, and all of the
// other solid-state controllers): each port is wired to ground through
// some kind of transistor (typically a MOSFET or Darlington).  When
// the output controller port is OFF, the transistor doesn't conduct to
// ground, so the port "floats"; when the port is ON, the transistor
// conducts to ground.
//
// The wiring to the DOF controller is a little tricky, because we're
// ALSO using those DOF controller connections to switch the physical
// strobe and beacon devices, which use relatively high power supply
// voltages (12V for each in my case) and high currents.  We can't
// expose the Trinket GPIO ports to those voltages or currents, so the
// Trinket needs to be isolated somehow.  I used optocouplers: the LED
// cathode of the optocoupler is wired to the DOF controller port,
// with the anode connected through a suitable resistor to + voltage,
// so that the DOF port turns on the optocoupler at the same time it
// fires the feedback device (beacon or strobe) that it's controlling.
// The opto collector is then connected to a Trinket GPIO input port,
// with a pull-up resistor to the Trinket's Vcc, and the emitter is
// connected to ground.  This makes the Trinket GPIO inputs read HIGH
// when the DOF port is OFF and LOW when ON.
//

// Trinket GPIO port numbers
const int R = 0;
const int G = 1;
const int B = 4;
const int Strobe = 3;
const int Beacons = 2;


// The light show is a program within a program.  I invented the
// small virtual machine language below as a simple way to express
// the pattern of flashing lights I wanted.  This works a lot like
// a machine language (a/k/a assembly language) for something like
// a primitive 8-bit processor: we have instructions to jump to a
// new program location, load and increment registers, and create
// loops using.  The registers are all brightness values for the LED
// color channels, plus a special loop counter register for the
// looping instructions.
//
// opcodes for the virtual machine language
const int JMP = 1;          // jump to fixed location
const int LDR = 2;          // load value into R brightness
const int LDG = 3;          // load value into G brightness
const int LDB = 4;          // load value into B brightness
const int LDW = 5;          // load value into White (== R G B set to same value)
const int INCR = 6;         // inc R brightness
const int INCG = 7;         // inc G brightness
const int INCB = 8;         // inc B brightness
const int INCW = 9;         // inc White (inc R G B by same amount)
const int DELAY = 10;       // delay (operand in ms)
const int HALT = 11;        // halt (effectively JMP <here>)
const int LOOP = 12;        // push program counter onto loop stack, push counter onto counter stack, load counter from operand
const int ENDLOOP = 13;     // dec counter by operand; if zero, pop old counter and pop loop stack, else jump to loop stack top

typedef struct op_t {
  char opcode;
  int operand;
} op_t;

// The virtual machine program to execute when all DOF devices are OFF
const op_t offProg[] = { 
  { LDW, 0 }, { HALT, 0 }
};

// The virtual machine program to execute when the strobe is on.
// This just flashes full-brightness white, effectively turning
// the fan into another strobe light.
//
// This takes precedence over the beacons, since the strobes fire
// in short bursts but the beacons tend to run for relatively long
// intervals.  We thus want the strobe to be able to interrupt a
// long-running beacon program, for a more dynamic effect.
const op_t strobeProg[] = {
  { LDW, 255 }, { HALT, 0 }
};

// The virtual machine program to execute when the beacons are on.
// This shows a complex series of short red/blue/white flashes.
// I wanted this to mimic the look of a crime scene, with a pile-up
// of police cruisers and ambulances all flashing their chaotic
// strobe patterns.  To keep things interesting, the program
// rotates through different patterns.
const op_t beaconProg[] = {
  // turn everything off initially
  { LDW, 0 },
  
  // outer loop - forever
  { LOOP, 1 }, 
  
    // flash red, flash blue 
    { LOOP, 1 },    
      // flash red, flash blue, then flash red/blue
      { LOOP, 1 },
        // flash red
        { LOOP, 6 },
          { LDR, 255 }, { DELAY, 40 },
          { LDR, 0 }, { DELAY, 60 },
        { ENDLOOP, 1 },
        // flash blue
        { LOOP, 6 },
          { LDB, 255 }, { DELAY, 40 },
          { LDB, 0 }, { DELAY, 60 },
        { ENDLOOP, 1 },
      { ENDLOOP, 1 },
    { ENDLOOP, 1 },
    
    // flash red/white/blue/white
    { LOOP, 1 },
      { LOOP, 6 },
        { LDR, 255 }, { DELAY, 40 },
        { LDR, 0 }, { DELAY, 60 },
      { ENDLOOP, 1 },
      { LOOP, 4 },
        { LDW, 255 }, { DELAY, 40 },
        { LDW, 0 }, { DELAY, 60 },
      { ENDLOOP, 1 },
      { LOOP, 6 },
        { LDB, 255 }, { DELAY, 40 },
        { LDB, 0 }, { DELAY, 60 },
      { ENDLOOP, 1 },
      { LOOP, 4 },
        { LDW, 255 }, { DELAY, 40 },
        { LDW, 0 }, { DELAY, 60 },
      { ENDLOOP, 1 },
    { ENDLOOP, 1 },
    
    // flash alternating red/blue
    { LOOP, 5 },
      { LDR, 255 }, { DELAY, 40 },
      { LDR, 0 }, { DELAY, 60 },
      { LDB, 255 }, { DELAY, 40 },
      { LDB, 0 }, { DELAY, 60 },
    { ENDLOOP, 1 },

  { ENDLOOP, 0 }  // loop forever
};

// Virtual machine program to run at system startup/power-on.  This
// is essentially a little test pattern that shows a series of color
// fades.  It makes it easy to check that the LEDs are working each
// time you turn on the machine, and it also creates a little visual
// interest while we're waiting for Windows to finish booting.
const op_t startupProg[] = {
  { LDW, 0 },
  { LOOP, 255 }, { INCR, 1 }, { DELAY, 2 }, { ENDLOOP, 1 },
  { LOOP, 255 }, { INCG, 1 }, { DELAY, 4 }, { ENDLOOP, 1 },
  { LOOP, 255 }, { INCR, -1 }, { DELAY, 2 }, { ENDLOOP, 1 },
  { LOOP, 255 }, { INCG, -1 }, { INCB, 1 }, { DELAY, 2 }, { ENDLOOP, 1 },
  { LOOP, 255 }, { INCR, 1 }, { DELAY, 4 }, { ENDLOOP, 1 },
  { LOOP, 255 }, { INCR, -1 }, { INCB, -1 }, { DELAY, 2 }, { ENDLOOP, 1 },
  { HALT, 1 }
};

// startup active; stays true until startup program halts
int startup = 1;

// current mode; 'prog' is the program for the current mode
int mode = 0;
const op_t *const progs[] = { startupProg, offProg, strobeProg, beaconProg };
const op_t *prog = startupProg;

// current program step
int pc = 0;

// loop stack
typedef struct stack_t {
  int loop;
  int cnt;
} stack_t;
const int MAX_STACK = 20;
stack_t stack[MAX_STACK];
int sp = 0;

// loop counter
int cnt = 0;

// halt mode start time
unsigned long halt_t0 = 0;

// current output brightness
int Rcur = 0, Gcur = 0, Bcur = 0;

void setup() {
  pinMode(R, OUTPUT);
  pinMode(G, OUTPUT);
  pinMode(B, OUTPUT);
  pinMode(Strobe, INPUT);
  pinMode(Beacons, INPUT);
}

void loop() {
  // get the current mode
  int newMode = 
      (startup ? 0 :
       digitalRead(Strobe) == LOW ? 2 :
       digitalRead(Beacons) == LOW ? 3 :
       1);
  
  if (newMode != mode)
  {
    mode = newMode;
    prog = progs[mode];
    pc = 0;
    sp = 0;
    cnt = 0;
  }

  // check for halt mode
  if (pc == -1)
  {
    // turn everything off if we've been halted too long (in case we're
    // stuck, to avoid overheating LEDs or MOSFETs)
    if (millis() - halt_t0 > 1000) {
      analogWrite(R, 0);
      analogWrite(G, 0);
      analogWrite(B, 0);
    }
    
    // do nothing more in halt mode
    return;
  }
  
  const op_t *op = &prog[pc++];
  switch (op->opcode)
  {
  case JMP:
    pc = op->operand;
    break;
    
  case LDR:
    Rcur = op->operand;
    break;
    
  case LDG:
    Gcur = op->operand;
    break;
    
  case LDB:
    Bcur = op->operand;
    break;
    
  case LDW:
    Rcur = Gcur = Bcur = op->operand;
    break;
    
  case INCR:
    Rcur += op->operand;
    break;
    
  case INCG:
    Gcur += op->operand;
    break;
    
  case INCB:
    Bcur += op->operand;
    break;
    
  case INCW:
    Rcur += op->operand;
    Gcur += op->operand;
    Bcur += op->operand;
    break;
    
  case DELAY:
    delay(op->operand);
    break;
    
  case HALT:
    // set halt mode
    pc = -1;
    
    // note halt mode start time
    halt_t0 = millis();
    
    // startup mode ends on the first halt
    startup = 0;
    break;
    
  case LOOP:
    if (sp + 1 < MAX_STACK) {
      stack[sp].loop = pc;
      stack[sp].cnt = cnt;
      ++sp;
    }
    cnt = op->operand;
    break;
    
  case ENDLOOP:
    cnt -= op->operand;
    if (sp > 0) {
      if (cnt == 0) {
        --sp;
        cnt = stack[sp].cnt;
      }
      else
        pc = stack[sp-1].loop;
    }
    break;
  }
  
  analogWrite(R, Rcur);
  analogWrite(G, Gcur);
  analogWrite(B, Bcur);
}

